/*
 * @(#)ReachableExcludes.java	1.1 98/03/06
 *
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package hat.model;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.BufferedReader;
import java.io.IOException;

import java.util.Hashtable;

/**
 * This represents a set of data members that should be excluded from the
 * reachable objects query.
 * This is useful to exclude observers from the
 * transitive closure of objects reachable from a given object, allowing
 * some kind of real determination of the "size" of that object.
 *
 * @version     1.1, 03/06/98
 * @author      Bill Foote
 */




public class ReachableExcludes {

    private File excludesFile;
    private long lastModified;
    private Hashtable methods;	// Hashtable<String, String>, used as a bag

    /**
     * Create a new ReachableExcludes over the given file.  The file will be
     * re-read whenever the timestamp changes.
     */
    public ReachableExcludes(File excludesFile) {
	this.excludesFile = excludesFile;
	readFile();
    }

    private void readFileIfNeeded() {
	if (excludesFile.lastModified() != lastModified) {
	    synchronized(this) {
		if (excludesFile.lastModified() != lastModified) {
		    readFile();
		}
	    }
	}
    }

    private void readFile() {
	long lm = excludesFile.lastModified();
	Hashtable m = new Hashtable();

	try {
	    BufferedReader r = new BufferedReader(new InputStreamReader(
				    new FileInputStream(excludesFile)));
	    
	    String method;
	    while ((method = r.readLine()) != null) {
		m.put(method, method);
	    }
	    lastModified = lm;
	    methods = m;	// We want this to be atomic
	} catch (IOException ex) {
	    System.out.println("Error reading " + excludesFile + ":  " + ex);
	}
    }

    /**
     * @return true iff the given field is on the histlist of excluded
     * 		fields.
     */
    public boolean isExcluded(String fieldName) {
	readFileIfNeeded();
	return methods.get(fieldName) != null;
    }
}
